@@ -2,9 +2,7 @@ from functools import partial, wraps
 from io import IOBase
 from typing import Any, Union, Sequence, Callable, Iterator
 import copy
-import hashlib
 import itertools
-import json
 import multiprocessing as mp
 import os
 import pickle
@@ -21,6 +19,8 @@ import sys
 import numpy as np
 import pandas as pd
 
+import joblib
+
 pj = os.path.join
 
 # defaults, globals
@@ -145,12 +145,10 @@ def pset_hash(
     #     >>> hash("12")
     #     4021864388667373027
     #
-    # We only try to hash a dict if it is json-serializable. Things which
-    # aren't are also not hashable in a reproducible fashion w/o spending much
-    # more work to solve the problem. We can't hash, say, the pickled byte
-    # string of some object (e.g. ``hash(pickle.dumps(obj))``), b/c that may
-    # contain a ref to its memory location which is not what we're interested
-    # in. Similarly, also using ``repr`` is not reproducible::
+    # We can't hash, say, the pickled byte string of some object (e.g.
+    # ``hash(pickle.dumps(obj))``), b/c that may contain a ref to its memory
+    # location which is not what we're interested in. Similarly, also using
+    # ``repr`` is not reproducible::
     #
     #     >>> class Foo:
     #     ...     pass
@@ -161,14 +159,11 @@ def pset_hash(
     #     '<__main__.Foo object at 0x7fcc732034c0>'
     #
     # even though for our purpose, we'd consider the two instances of ``Foo`` to
-    # be the same. Better be safe and return NaN in that case.
+    # be the same.
     #
     # The same observations have been also made elsewhere [1,2]. Esp. [2]
     # points to [3] which in turn mentions joblib.hashing.hash(). It's code
-    # shows how complex the problem is. We may try joblib's solution in the
-    # future should this feature become so important that we need to require
-    # hashes to work or else fail. ATM we treat pset hashes on a best effort
-    # basis b/c we don't have strong use cases for them just yet.
+    # shows how complex the problem is, but so far this is our best bet.
     #
     # [1] https://death.andgravity.com/stable-hashing
     # [2] https://ourpython.com/python/deterministic-recursive-hashing-in-python
@@ -179,9 +174,11 @@ def pset_hash(
             _dct = {kk: dct[kk] for kk in keys}
         else:
             _dct = dct
-        h = hashlib.new(method)
-        h.update(json.dumps(_dct, sort_keys=True).encode())
-        return h.hexdigest()
+        return joblib.hash(_dct, hash_name=method)
+    # joblib can hash "anything" so we didn't come up with an input that
+    # actually fails to hash. As such, TypeError is just a guess here. But
+    # still we don't catch ValueError raised when an invalid hash_name is
+    # passed (anything other than md5 or sha1).
     except TypeError as ex:
         if raise_error:
             raise PsweepHashError(
@@ -370,19 +367,31 @@ def df_to_json(df: pd.DataFrame, **kwds) -> str:
     return df.to_json(**kwds)
 
 
-def df_write(df: pd.DataFrame, fn: str, fmt="pickle", **kwds) -> None:
+def df_write(fn: str, df: pd.DataFrame, fmt="pickle", **kwds) -> None:
     """Write DataFrame to disk.
 
     Parameters
     ----------
-    df : DataFrame
     fn :
         filename
+    df : DataFrame
     fmt :
         ``{'pickle', 'json'}``
     kwds :
         passed to ``pickle.dump()`` or :func:`df_to_json`
     """
+    _swapped_args_err_msg = (
+        "Deprecated API in df_write(), use df_write(filename, df)"
+    )
+    if isinstance(fn, pd.DataFrame) or isinstance(df, str):
+        if isinstance(fn, pd.DataFrame) and isinstance(df, str):
+            warnings.warn(_swapped_args_err_msg, DeprecationWarning)
+            tmp_fn = df
+            df = fn
+            fn = tmp_fn
+        else:
+            raise ValueError(_swapped_args_err_msg)
+
     makedirs(os.path.dirname(fn))
     if fmt == "pickle":
         with open(fn, "wb") as fd:
@@ -390,7 +399,7 @@ def df_write(df: pd.DataFrame, fn: str, fmt="pickle", **kwds) -> None:
     elif fmt == "json":
         df_to_json(df, path_or_buf=fn, **kwds)
     else:
-        raise Exception("unknown fmt: {}".format(fmt))
+        raise ValueError("unknown fmt: {}".format(fmt))
 
 
 def df_read(fn: str, fmt="pickle", **kwds):
@@ -404,7 +413,7 @@ def df_read(fn: str, fmt="pickle", **kwds):
             fn, precise_float=True, orient=orient, **kwds
         )
     else:
-        raise Exception("unknown fmt: {}".format(fmt))
+        raise ValueError("unknown fmt: {}".format(fmt))
 
 
 def df_print(
@@ -777,14 +786,13 @@ def worker_wrapper(
     assert calc_dir is not None
     pset_id = str(uuid.uuid4())
     _pset = copy.deepcopy(pset)
-    hash_alg = PSET_HASH_ALG
     time_start = pd.Timestamp(time.time(), unit=PANDAS_TIME_UNIT)
     update = {
         "_run_id": run_id,
         "_pset_id": pset_id,
         "_calc_dir": calc_dir,
         "_time_utc": time_start,
-        f"_pset_{hash_alg}": pset_hash(pset, hash_alg, raise_error=False),
+        "_pset_hash": pset_hash(pset, PSET_HASH_ALG, raise_error=False),
         "_pset_seq": pset_seq,
         "_run_seq": run_seq,
     }
@@ -802,7 +810,7 @@ def worker_wrapper(
     df_row = pd.DataFrame([_pset])
     if tmpsave:
         fn = pj(calc_dir, "tmpsave", run_id, pset_id + ".pk")
-        df_write(df_row, fn)
+        df_write(fn, df_row)
     return df_row
 
 
@@ -945,7 +953,7 @@ def run_local(
         df = pd.concat((df, df_row), sort=False, ignore_index=True)
 
     if save:
-        df_write(df, database_fn)
+        df_write(database_fn, df)
 
     git_exit(git, df)
 
@@ -1028,6 +1036,7 @@ def prep_batch(
     machine_templ_dir: str = "templates/machines",
     git: bool = False,
     backup: bool = False,
+    write_pset: bool = False,
 ) -> pd.DataFrame:
 
     """
@@ -1045,6 +1054,8 @@ def prep_batch(
                 pj(calc_dir, pset["_pset_id"], template.targetname),
                 template.fill(pset),
             )
+            if write_pset:
+                pickle_write(pj(calc_dir, pset["_pset_id"], "pset.pk"), pset)
         return {}
 
     df = run_local(worker, params, calc_dir=calc_dir, backup=backup, git=False)
@@ -1057,7 +1068,7 @@ def prep_batch(
             if msk.any():
                 txt += "\n"
             txt += "\n".join(
-                f"{pfx}cd {pset_id}; {machine.subcmd} {machine.template.targetname}; cd $here  # run_seq={run_seq} pset_seq={pset_seq}"
+                f"{pfx}cd $here/{pset_id}; {machine.subcmd} {machine.template.targetname}  # run_seq={run_seq} pset_seq={pset_seq}"
                 for pset_id, pset_seq, run_seq in zip(
                     df[msk]._pset_id.values,
                     df[msk]._pset_seq.values,
@@ -1066,7 +1077,7 @@ def prep_batch(
             )
         file_write(
             f"{calc_dir}/run_{machine.name}.sh",
-            f"#!/bin/sh\n\nhere=$(pwd)\n{txt}\n",
+            f"#!/bin/sh\n\nhere=$(readlink -f $(dirname $0))\n{txt}\n",
         )
 
     git_exit(git, df)
